/***************************************************************************
 * Factor5j v0.99 - June. 23th 2018                                        *
 * by Luigi Morelli - l.morelli@moreware.org                               *
 *                                                                         *
 * Mersenne and double Mersenne numbers factorization program              *
 * Special version for k range <= 250000 and 2^32 < exponent < 2^62        *
 * Doesn't get input from either file or command line                      *
 * Uses INI file to guard status                                           *
 *                                                                         *
 * Works without GMP! New number-theoretical routines                      *
 *                                                                         *
 * compile with: gcc factor5j.c -O3 -std=c99 -ofactor5j                    *
 *                                                                         *
 * run with    : ./factor5j                                                *
 *                                                                         *
 *                                                                         *
 * Files produced:                                                         *
 * - results.txt: this file contains the results for a complete            *
 *   run, and the eventual factor(s) found, in different formats:          *
 *    1 - Factor5 original format                                          *
 *    2 - Only the string "Mxxx has a factor: yyy" is printed              *
 *    4 - the line exponent,k is printed for each found exponent           *
 *    8 - simil-pfgw: xxxxxx | 2^yyyyy-1                                   *
 *    Being bitmapped, most combinations work together as well             *
 *                                                                         *
 * -factor5j.ini: this file contains information about the run:            *
 *   Exponent, exponent_f, Kstart, Kend, out_flag, stop_after_factor       *
 *                                                                         *
 * Being an experimental tool, no strong argument checking is done         *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <inttypes.h>
#include <unistd.h>
#include <getopt.h>
#include <ctype.h>
// #include <gmp.h>
// #include <math.h>
// #include <time.h>
// #include <pthread.h>

//#define DEBUG

#define TRUE		1  	// true
#define FALSE		0  	//false
#define PRIME 		1
#define COMPOSITE 	0

#define MAXPRIMES	13000				// from 1.300.000 to 15.000.000 depending on exponent and factoring bit depth
#define MAXK		250000	 			// 250000 - 272272
#define INIPARAMS	8					// number of parameters in the INI file


#define version "1.01 - August 3, 2018 - AMD64/Win64 compile"

char *results_filename = "results.txt";
char *INI_filename     = "factor5j.ini";
int sieve_folder [MAXPRIMES];			// Needed for Eratosthenes' sieve
int start_gap[1600];					// Holds the primes in 1.. MAXPRIMES

// const uint64_t bases64[] = {2, 325, 9375, 28178, 450775, 9780504, 1795265022};	// original W.Izykowski Miller-Rabin witnesses bases
const uint64_t bases64[] = {2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37};	
										// for deterministic Miller-Rabin tests up to exponents = 2^64 (see Excel table)
int MR_bit_limit = 6;					// the code has been actually hard-limited to six witnessess bases, or to exponent  3,474,749,660,383

int num_primes = 0;
int max_primes;

// some globals
	uint64_t exponent    = 4340000000;		// The exponent of the Mersenne number to test                 
	uint64_t exponent_f  = 1;				// The range of exponents to be testeed in a run
	uint64_t Kstart      = 1;				// The starting range of K
	uint64_t Kend        = 250000;			// The starting range of K
	int      out_flag    = 7;				// bitmapped : 1 = current factor - 2 = has a factor - 4 = exponent,k
	int      stop_factor = 0;				// 0 = complete range - 1 = stop after a factor is found
	int      vid_verbose = 1;				// The program will print the results on the screen 
	int      last_line   = 1;				// The program will print a line with the stats on the screen 

	uint64_t k, maxk, factor, twop;
	int factors_found = 0;


// Function prototyping
int Is_prime(uint64_t);
uint64_t next_prime(uint64_t);
static inline uint64_t powmod(uint64_t, uint64_t, uint64_t);
static inline int MR_test(uint64_t);


/******************************************************************************************************************************
 * SERVICE FUNCTIONS                                                                                                          *
 ******************************************************************************************************************************/

/********************************************************
 * split(char *)                                        *
 *                                                      *
 * INPUT  : A string to parse                           *
 * OUTPUT : a unsigned long int                         *
 *                                                      *
 * ACTION : takes the string in input and splits it     *
 *          after the '=' sign, returning the leftover  *
 *          converted as a usigned long int             *
 *          Need some error-checking                    *
 *                                                      *
 * DEFAULT VALUES: NONE                                 *
 *                                                      *
 ********************************************************/
uint64_t split(char* s) {	
	char *dest;
  
	dest = strstr(s, "=");
	return atoll(++dest);

}


/********************************************************
 * bitcount()                                           *
 *                                                      *
 * INPUT  : The number to count bits (uint64_t)         *
 * OUTPUT : Number of significant bits                  *
 *                                                      *
 * ACTION : Defines the most significant bit            *
 *          counting down from 64 until a 1 is found    *
 *                                                      *
 * DEFAULT VALUES: NONE                                 *
 *                                                      *
 ********************************************************/
int bitcount(uint64_t a) {
    int i; 
    for(i=63; i>=0; i--)
          if(a&(1lu<<i)) return i;
	return 0;
}

/********************************************************
 * parseINI()                                           *
 *                                                      *
 * INPUT  : None                                        *
 * OUTPUT : None                                        *
 *                                                      *
 * ACTION : Looks for a INI file                        *
 *          If the file is found, parses the variables  *
 *          Else stops complaining                      *
 *          there must be NO space btw key and value!   *
 *          Ex: exponent=4294967311                     *
 *                                                      *
 * DEFAULT VALUES:                                      *
 *          exponent    = 4,340,000,000                 *
 *          exponent_f  = 1			                    *
 *          Kstart      = 1                             *
 *          Kend        = 250000                        *
 *          out_flag    = 7                             *
 *          stop_factor = 0                             *
 *          vid_verbose = 1                             *
 *          last_line   = 1                             *
 *                                                      *
 ********************************************************/
void parseINI() {

	FILE *fd;
	char  str[20];
	int i = 0;

	// Is there a INI file?
	fd = fopen(INI_filename, "r");

	if (fd) {
		for (i = 0; i < INIPARAMS; i++) {
			int a = fscanf(fd, "%s", str); 
			if (a);

			// Unpack values...	
			switch (i)
			{
				case 0  : exponent    =      split(str); if (exponent   < 1) {printf("Error parsing INI file!\n"); exit(-1);}; break;
				case 1  : exponent_f  = 	 split(str); if (exponent_f < exponent) {printf("Error parsing INI file!\n"); exit(-2);}; break;
				case 2  : Kstart      =      split(str); if (Kstart	    < 1)  {printf("Error parsing INI file!\n"); exit(-3);}; break;
				case 3  : Kend	      =      split(str); if (Kend		< 1)  {printf("Error parsing INI file!\n"); exit(-3);}; break;
				case 4  : out_flag    = (int)split(str); if ((out_flag  < 1) || (out_flag > 7)) {printf("Error parsing INI file!\n"); exit(-4);}; break;
				case 5  : stop_factor = (int)split(str); break;
				case 6  : vid_verbose = (int)split(str); break;
				case 7  : last_line   = (int)split(str); if ((last_line   < 0) || (last_line > 1)) {printf("Error parsing INI file!\n"); exit(-5);};break;
			}
		}			
		fclose(fd);
	} 

	int nbits = bitcount(exponent); 
	switch (nbits) {
		case 45: maxk = 0x20000; break;
		case 46: maxk = 0x10000; break;
		case 47: maxk = 0x8000;  break;
		case 48: maxk = 0x4000;  break;
		case 49: maxk = 0x2000;  break;
		case 50: maxk = 0x1000;  break;
		case 51: maxk = 0x800;   break;
		case 52: maxk = 0x400;   break;
		case 53: maxk = 0x200;   break;
		case 54: maxk = 0x100;   break;
		case 55: maxk = 0x80;    break;
		case 56: maxk = 0x40;    break;
		case 57: maxk = 0x20;    break;
		case 58: maxk = 0x10;    break;
		case 59: maxk = 0x8;     break;
		case 60: maxk = 0x4;     break;
		case 61: maxk = 0x2;     break;
		case 62: maxk = 0x1;     break;
		default: maxk = MAXK; 	 break;
	}

	Kend = Kstart + maxk;
	return;
}



/******************************************************************************************************************************
 * NUMBER THEORY FUNCTIONS                                                                                                    *
 ******************************************************************************************************************************/

/********************************************************
 * mulmod(int, int, int)                                *
 *                                                      *
 * INPUT  : a, b, modulus                               *
 * OUTPUT : modulo                                      *
 *                                                      *
 * ACTION : Performs the modular multiplication         *
 *          using the Montgomery algorithm              *
 *                                                      *
 * DEFAULT VALUES: NONE                                 *
 *                                                      *
 ********************************************************/
static inline uint64_t mulmod64(uint64_t a, uint64_t b, uint64_t n)
{
	uint64_t d;
	uint64_t unused; // dummy output, unused, to tell GCC that RAX register is modified by this snippet
	__asm__ ("mulq %3\n\t"
		 "divq %4"
		 :"=a"(unused), "=&d"(d)
		 :"a"(a), "rm"(b), "rm"(n)
		 :"cc");
	return d;
}

/********************************************************
 * powmod(int, int, int)                                *
 *                                                      *
 * INPUT  : Base, exponent, modulus                     *
 * OUTPUT : 1 if divides                                *
 *                                                      *
 * ACTION : Performs the modular exponentiation         *
 *          using the right-to-left algorithm           *
 *          and returns 1 if m is a factor of b^e       *
 *                                                      *
 * DEFAULT VALUES: NONE                                 *
 *                                                      *
 ********************************************************/

static inline uint64_t powmod(uint64_t a, uint64_t b, uint64_t n)
{
	uint64_t d=1, A=a;

	do {
		if (b&1)
			d=mulmod64(d, A, n);
		A=mulmod64(A, A, n);
	} while (b>>=1);

	return (uint64_t)d;
}


/********************************************************
 * next_prime(uint64)                                   *
 *                                                      *
 * INPUT  : starting prime                              *
 * OUTPUT : prime next the input                        *
 *                                                      *
 * ACTION : Finds the prime next to the one in input    *
 *          calling the is_prime() function             *
 *          and returns the next prime                  *
 *                                                      *
 * DEFAULT VALUES: NONE                                 *
 *                                                      *
 ********************************************************/

uint64_t next_prime(uint64_t p)
{
	do {
		p = p + 2;
	} while (Is_prime(p) == 0);

	return (uint64_t)p;
}


/********************************************************
 * Is_prime(uint64)                                     *
 *                                                      *
 * INPUT  : prime candidate                             *
 * OUTPUT : 1 = prime - 0 = composite                   *
 *                                                      *
 * ACTION : tests the candidate for primality           *
 *          using deterministic Miller-Rabin algorithm  *
 *          for numbers < 2^64                          *
 *                                                      *
 * DEFAULT VALUES: NONE                                 *
 *                                                      *
 ********************************************************/

int Is_prime(uint64_t p)
{
	// lista di primi da testare
	int threshold = 18;  //0x1000

	// test for small ints < threshold
	for (int i = 2; i < threshold; i++) 	
		if ((p % start_gap[i]) == 0) return 0;  
	
	if (MR_test(p) == 0) return 0;

	return 1;
}

/********************************************************
 * MR_test(uint64, it, uint64)                          *
 *                                                      *
 * INPUT  : bases, counter for bases, factor_to_test    *
 * OUTPUT : 1 = prime - 0 = composite                   *
 *                                                      *
 * ACTION : Deterministic Miller-Rabin algorithm        *
 *          (from an idea of W.Izykowski)               *
 *          for numbers < 2^64                          *
 *                                                      *
 * DEFAULT VALUES: NONE                                 *
 *                                                      *
 ********************************************************/
static inline int MR_test(uint64_t n)
{
	uint64_t u=n-1;

	int t = __builtin_ctzll(u), j;
	u >>= t;

	for (j=0; j<MR_bit_limit; j++) {
		uint64_t a = bases64[j], x;
		int i;

		if (a >= n) a %= n;

		if (a == 0) continue;

		x = powmod(a, u, n);

		if (x == 1 || x == n-1) continue;

		for (i=1; i<t; i++) {
			x=mulmod64(x, x, n);
			if (x == 1)   return 0;
			if (x == n-1) break;
		}

		// if we didn't break, the number is composite
		if (i == t) return 0;
	}

	return 1;
}



/******************************************************************************************************************************
 * WORK FUNCTIONS                                                                                                             *
 ******************************************************************************************************************************/

/********************************************************
 * sieve()                                              *
 *                                                      *
 * INPUT  : None                                        *
 * OUTPUT : None                                        *
 *                                                      *
 * ACTION : Performs the Eratosthenes' sieve            *
 *            and set up the start_gap structure        *
 *            to use appropriate residual classes       *
 *                                                      *
 * DEFAULT VALUES: NONE                                 *
 *                                                      *
 ********************************************************/
void sieve()
{
	int i = 0;
	int j;

	num_primes = 0;

	// Inizialization of variables
	memset( sieve_folder, 0, sizeof( sieve_folder ) );

	i = 2;
	while ( i < MAXPRIMES )
	{
		j = i;
		if ( sieve_folder[j] == 0 )
		{
			if ( i != exponent )
			{
				start_gap[num_primes] = i; 

				num_primes++;
			}

			while ( j < MAXPRIMES - i )
			{
				j = j + i;
				sieve_folder[j] = 1;
			}
		}
		i++;
	} 	

	return;
}


/********************************************************
 * factorize()                                          *
 *                                                      *
 * INPUT  : None                                        *
 * OUTPUT : None                                        *
 *                                                      *
 * ACTION : Here the sieve and powmod job is done       *
 *          the function is a parallel thread           *
 *                                                      *
 * DEFAULT VALUES: NONE                                 *
 *                                                      *
 ********************************************************/
void factorize()
{	
	FILE * fd;

	uint64_t i, Mf;

	for (i = Kstart; i < Kend; i++) {
		Mf = twop * i + 1;

		uint64_t j = Mf % 120;
		if ( ( j == 1   ) || ( j == 7   ) || ( j == 17  ) || ( j == 23 ) || 
			 ( j == 31  ) || ( j == 41  ) || ( j == 47  ) || ( j == 49 ) ||
			 ( j == 119 ) || ( j == 113 ) || ( j == 103 ) || ( j == 97 ) || 
			 ( j == 89  ) || ( j == 79  ) || ( j == 73  ) || ( j == 71 )
		   ) {
			for (int j = 4; j < 32; j++) {
				if (Mf % start_gap[j] == 0)
					goto COMP;
			}
			if (powmod(2, exponent, Mf) == 1) {
				if (vid_verbose == 1) 
					printf ("%30lu | 2^%lu-1 -  k = %20lu\n", Mf, exponent, i);
				
				if ((out_flag &1) || (out_flag & 2)) {
					fd = fopen(results_filename, "a");
					if (fd) {
						fprintf(fd, "M%lu has a factor: %lu\n", exponent, Mf );
						fclose(fd);
					}
				}

				if ((out_flag & 4)) {
					fd = fopen(results_filename, "a");
					if (fd) {
						fprintf(fd, "%lu,%lu\n", exponent, i );
						fclose(fd);
					}
				}

				factors_found++;
				
				if (stop_factor == 0) break;
			}
		}
		COMP:;
	}
}


/******************************************************************************************************************************
 * MAIN                                                                                                                       *
 ******************************************************************************************************************************/

/********************************************************
 * main()                                               *
 *                                                      *
 * INPUT  : None                                        *
 * OUTPUT : return code                                 *
 *                                                      *
 * ACTION : Looks for factors of Mersenne numbers       *
 *          using nice heuristics and multithreading    *
 *          Values taken from INI file or automatic     *
 *                                                      *
 ********************************************************/
int main ( int argc, char *argv[] ) 
{
	int total_factors_found = 0;
	int num_exp_tested		= 0;
	uint64_t first_exponent;

	FILE * fd;
	FILE * fd1;
	
	/******************************************************** 
	 * Here the default parameters are:						*
	 * DEFAULT VALUES:                                      *
	 *          exponent    = 4,340,000,000                 *
	 *          exponent_f  = 4,340,000,000                 *
	 *          Kstart      = 1                             *
	 *          Kend        = 250000                        *
	 *          out_flag    = 7                             *
	 *          stop_factor = 0                             *
	 *          vid_verbose = 1                             *
	 *          last_line   = 1                             *
	 *                                                      *
	 ********************************************************/	
	 
	// Now let's try and parse the parameters inside the INi file...
	parseINI();	 
	 	
	/********************************************************
	 * parseCL()                                            *
	 *                                                      *
	 * ACTION : Extract command line parameters (if any)    *
	 *                                                      *
	 * DEFAULT VALUES: NONE                                 *
	 *                                                      *
	 ********************************************************/
	int c, ck = -1;
	char s_exp[20];
	int is_worktodo = FALSE;
	opterr = 0;

	while ((c = getopt (argc, argv, "f:e:E:k:K:")) != -1)  {
		switch (c)
		{
			case 'e': exponent   = atoll(optarg); break;
			case 'E': exponent_f = atoll(optarg); break;
			case 'k': Kstart     = atoll(optarg); break;
			case 'K': Kend       = atoll(optarg); break;
			case 'f': fd1 = fopen(optarg, "r");
						if (!fd1) {
							// printf("Cannot open file worktodo.txt\n");
							exit(0);
						}
						else {		
							// printf("Working witn file\n");
							is_worktodo = TRUE;
							ck = fscanf(fd1, "%s", s_exp);
							exponent = atoll(s_exp);							
						}
						break;
			case '?':
				if (isprint (optopt))
					fprintf (stderr, "Unknown option `-%c'.\n", optopt);
				else
					fprintf (stderr,"Unknown option character `\\x%x'.\n", optopt);
			default:exit (-10);
		}
	}
	
	// Parameters check
	if (is_worktodo == FALSE) {
		if (exponent > exponent_f) {
			printf("The starting exponent must be smaller than the final exponent!\n");
			exit(0);
		}
	}
	
	if (Kstart > Kend) {
		printf("The starting K must be smaller than the final K!\n");
		exit(0);
	}
		
	sieve();

	first_exponent = exponent;
	
	// Test the exponent primality
	if ((exponent % 2) == 0) 
		exponent++;
		
	if (!Is_prime(exponent))
		exponent = next_prime(exponent);

	// cycle over exponent_f		
	do {
		// Set up values
		factors_found = 0;
		k = Kstart;
		twop = 2*exponent;

		if (out_flag & 1) {
			fd = fopen(results_filename, "a");
			if (fd) {
				fprintf(fd, "Trial-factoring M%lu in K[%lu, %lu]\n", exponent, Kstart, Kend );
				fclose(fd);
			}
		}

		factorize();

		//fwrite_stats();
		if (out_flag & 1) {
			fd = fopen(results_filename, "a");
			if (fd) {
				fprintf(fd, "M%lu has %i factors in K[%lu, %lu].\n", exponent, factors_found, Kstart, Kend );
				fclose(fd);
			}
		} 
		
		total_factors_found = total_factors_found + factors_found;

		num_exp_tested++;
		if (is_worktodo == FALSE) 
			exponent = next_prime(exponent);
		else
		{
			ck = fscanf(fd1, "%s", s_exp);
			exponent = atoll(s_exp);
		}
	} while ((exponent_f > exponent) || (ck != -1));
	
	if (is_worktodo == TRUE) {
		fclose (fd1);
		exponent_f = exponent;
	}

	if(last_line == 1) printf("\nTested %u exponents over N[%lu..%lu] and K[%lu..%lu] - Found %i factors - Ratio: %6.3f %%\n", num_exp_tested, first_exponent, exponent_f, Kstart, Kend, total_factors_found, 100*(float)total_factors_found/num_exp_tested);
	return 0;
}
